/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#pragma once

#include "types.h"
#include "service.h"
#include <util/task/types.h>
#include <util/db/bdb/storage.h>
#include <boost/bind.hpp>
#include <boost/signals2.hpp>
#include <boost/function.hpp>
#include <string>
#include <vector>

#define RELEASE_SERVICE(srv) { if (srv) { srv->release(); srv = NULL; }}

namespace zapper {
namespace plugin {

class Plugin;
class Service;
typedef boost::signals2::signal<void (Service *, bool isRegistered)> OnRegisterChanged;
typedef util::storage::bdb::Storage StorageImpl;

/**
 * Esta clase tiene como tarea contener a todos los @em Servicios y @em Plugins, administrando su estado.
 */
class PluginManager {
public:
	explicit PluginManager( util::task::Dispatcher *disp );
	virtual ~PluginManager();

	//	Add plugin(s)
	void add( Plugin *plugin );
	//void add( const std::string &dsoPath );
	bool isPresent( const std::string &plugin ) const;
	bool isEnabled( const std::string &plugin ) const;
	void enable( const std::string &plugin, bool state );

	//	Plugins initialization
	void initialize( const std::string &dbDir );
	void finalize();
	bool isInitialized() const;

	//	Start/Stop services
	void start();
	void stop();
	bool isStarted() const;

	//	Standby
	bool inStandby() const;
	void standby();
	void wakeup();

	//	Configuration
	void resetConfig();

	//	Register/Unregister services
	const std::vector<Service *> &services() const;
	bool registerService( Service *srv );
	void unregisterService( const std::string &instanceName );
	std::vector<std::string> findInstances( const std::string &className ) const;
	template<typename T> T *findService( const std::string &instanceName ) const;
	template<typename T> T *addRef( const std::string &instanceName ) const;
	template<typename T> void runMethod( const std::string &instanceName, const boost::function<void (T *)> &method );
	bool isServiceRegistered( const std::string &instanceName ) const;
	bool isServiceStarted( const std::string &instanceName ) const;
	Service *getService( const std::string &instanceName ) const;
	OnRegisterChanged &onRegisterChanged();

	//	Service notifications
	void onServiceChanged( Service *srv );

	// Save/load configuration to/from storage
	template<typename T>
	void save( const std::string &key, const T &value );

	template<typename T>
	void load( const std::string &key, T &value );

protected:
	//	Types
	typedef std::vector<Service *> Services;

	//	Aux
	void showStatus();
	void start( Service *srv );
	void stop( Service *srv );
	void cleanServices();
	void unregisterService( Service *srv );
	Plugin *findPlugin( const std::string &name ) const;
	template<typename T> void execute( const std::string &name, const boost::function<void (T *)> &method );
	void post( const util::task::Type &task );

private:
	util::task::Dispatcher *_tasks;
	plugin::status::type _status;
	StorageImpl *_storage;
	std::vector<Plugin *> _plugins;
	Services _services;
	OnRegisterChanged _onRegisterChanged;
};

/**
 * @param name Nombre de la instancia del servicio.
 * @return Puntero al servicio buscado o NULL en caso de que no exista en la colección de servicios del objeto PluginManager.
 */
template<typename T> T *PluginManager::findService( const std::string &name ) const {
	Service *srv = getService( name );
	if (srv) {
		return srv->get<T>();
	}
	return NULL;
}

/**
 * Informa al servicio que tiene una referencia más.
 * @param name Nombre de la instancia del servicio.
 * @return Puntero al servicio al cuál se le ha agregado una referencia.
 */
template<typename T> T *PluginManager::addRef( const std::string &name ) const {
	T *srv = findService<T>( name );
	if (srv) {
		srv->addRef();
	}
	return srv;
}

/**
 * Agrega una tarea para ser ejecutada en loop principal.
 * @param name Nombre de la instancia del servicio.
 * @param method Función a ser ejecutada.
 */
template<typename T>
void PluginManager::runMethod( const std::string &name, const boost::function<void (T *)> &method ) {
	BOOST_ASSERT(!method.empty());	
	boost::function<void (void)> fnc = boost::bind(&PluginManager::execute<T>,this,name,method);
	post( fnc );
}

/**
 * Ejecuta una tarea sobre el servicio especificado.
 * @param name Nombre de la instancia del servicio.
 * @param method Función a ser ejecutada.
 */
template<typename T>
void PluginManager::execute( const std::string &name, const boost::function<void (T *)> &method ) {
	BOOST_ASSERT(!method.empty());
	T *srv = findService<T>( name );
	if (srv && srv->isStarted()) {
		method( srv );
	}
}

/**
 * Almacena en la base de datos un valor.
 * @param key Clave asociada al valor.
 * @param value Valor que se desea almacenar.
 */
template<typename T>
void PluginManager::save( const std::string &key, const T &value ) {
	BOOST_ASSERT(_storage);
	_storage->save( key, value );
}

/**
 * Recupera un valor de la base de datos.
 * @param key Clave asociada al valor.
 * @param value Valor que se desea recuperar.
 */
template<typename T>
void PluginManager::load( const std::string &key, T &value ) {
	BOOST_ASSERT(_storage);
	_storage->load( key, value );
}

}
}

