/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#pragma once

#include <mpegparser/types.h>
#include <util/keydefs.h>
#include <boost/function.hpp>
#include <vector>
#include <queue>

namespace tuner {
	class Provider;
}

namespace canvas {
	class System;
	namespace event {
		union NotifyStruct;
		typedef union NotifyStruct Notify;
	}
}

namespace zapper {

namespace plugin {
	class PluginManager;
	class Plugin;
}

namespace input {
	class InputService;
}

namespace audio {
	class Mixer;
}

namespace display {
	class DisplayService;
}

namespace player {
	class Player;
}

namespace middleware {
	class ApplicationController;
}

class MountService;

/**
 * Es el punto de inicio para que los servicios estén disponibles.
 */
class Zapper {
public:
	explicit Zapper( const std::string &name );
	virtual ~Zapper();

	//	Initialize methods
	bool initialize();
	void finalize();

	//	Start/stop methods
	void start();
	void stop();

	//	Loop method
	void run();
	void exit( int ret=0 );
	int exitCode();

	//	Add/Enable plugins
	void addPlugin( plugin::Plugin *plugin );
	plugin::PluginManager *mgr() const;

	//	Restore to default configuration
	void resetConfig();

	//	Version
	std::string getVersion() const;
	std::string getPlatformVersion() const;

	//	RAM disk
	const std::string &ramDisk() const;

	//	Canvas
	canvas::System *sys() const;

	std::string name() const;

	/**
	 * Crea un objeto zapper.
	 * @param use Nombre del zapper.
	 * @return Una instancia de @c Zapper.
	 */
	static Zapper *create( const std::string &use );

protected:
	//	Hardware initialization
	virtual bool initHardware();
	virtual void finHardware();
	virtual bool initCanvas();

	//	Start/Stop
	virtual void onStart();
	virtual void onStop();

	//	Services registration
	virtual void registerPlugins();
	virtual void registerServices( plugin::PluginManager *mgr );

	//	Standby
	virtual bool supportStandby() const;
	virtual void doStandby();
	virtual void doWakeup();
	void standby();
	void wakeup();
	void onPowerKey( util::key::type key, bool isUp );	

	//	Instance creation
	virtual tuner::Provider *createProvider() const;
	virtual canvas::System *createCanvasSystem() const;	
	virtual display::DisplayService *createDisplay() const;
	virtual bool haveMixer() const;
	virtual audio::Mixer *createMixer() const;
	virtual player::Player *createPlayer() const;
	virtual MountService *createMount() const;
	virtual middleware::ApplicationController *createMiddleware() const;

	//	Enqueue task into zapper thread
	void onSystemEvent( canvas::event::Notify *notify );

	//	Paths
	bool setupPaths( std::string &db );

	//	Version
	int getMinVersion() const;

private:
	int _exitCode;
	std::string _name;
	canvas::System *_sys;
	plugin::PluginManager *_mgr;
	input::InputService *_input;
	std::string _ramDisk;
	bool _initialized;
	bool _shutdown;
};

}

